class AdminEventManager extends SystemListManager
	@is_allowed: ->
		ez5.session.hasSystemRight("root", "plugin.eventmanager.manager")

	@group: ->
		"zze_administration"

	@submenu: ->
		"administration"

	@path: ->
		["eventmanager"]

	@label: ->
		"admin.event.manager.menu.label"

	initOpts: (opts) ->
		if ez5.version("6")
			@__newApi = true
		else
			@__newApi = false
		super(opts)

	getTreeOpts: ->
		opts = super()

		opts.filterManager = ->

		filter = new CUI.ListViewHeaderRow()

		@__sortColumns = []

		columnClasses = []
		columnKeys = ["_id", "timestamp", "user_generated_displayname", "type", "type_key"]

		for key in columnKeys
			filter.addColumn(@getHeaderColumn(key))
			columnClasses.push("ez5-admin-event-manager-column-#{key}")

		opts.colClasses = columnClasses
		opts.extra_header = filter
		@__sort = "_id.DESC"
		@__updateSortButtons()
		opts.cols = ["auto","auto","auto","auto", "auto"] #,"auto"]
		opts.selectableRows = "multiple"

		opts

	loadInternal: ->
		super()
		@__layout.replace(
			new CUI.SimplePane
				class: "cui-pane--window"
				title: $$("admin.event.manager.title")
				content: @tree.render()
		, "center")
		return @

	readOpts: ->
		super()
		@__filter_data = {}

	getRightPane: ->
		content: {}
		flexHandle:
			label: @getMenuTitle()
			state_name: "admin-manager-right"
			hidden: false

	getContainer: ->
		@__layout.right()

	getEvents: ->
		events = ez5.session.getDefaults().server.events?.searchable_events.slice(0) or []
		events.sort (a, b) ->
			a.localeCompare(b)

	getTreePaneContent: ->
		groupOptions = []
		groupData = {}

		@__searchIdInput = new CUI.Input
			maximize_horizontal: true
			name: "id"
			placeholder: if @__newApi then $$("admin.event.manager.form.id.placeholder")
			form:
				label: $$("admin.event.manager.form.id")

		userTypesOptions = []

		@__filter_form = new CUI.Form
			data: @__filter_data
			padded: true
			onDataChanged: @__enableDisableFormFields
			fields: [
				@__searchIdInput
			,
				@getSelectFilter("type", @getEvents())
			,
				maximize_horizontal: true
				type: CUI.DateTime
				ui: "admin-event-manager-form-date-from-datetime"
				undo_and_changed_support: false
				form:
					label: $$("admin.event.manager.form.date_from")
				name: "from"
				display_type: "short"
				input_types: ["date_time"]
			,
				maximize_horizontal: true
				type: CUI.DateTime
				ui: "admin-event-manager-form-date-to-datetime"
				undo_and_changed_support: false
				form:
					label: $$("admin.event.manager.form.date_to")
				name: "to"
				display_type: "short"
				input_types: ["date_time"]
			,
				if ez5.version("6.7")
					maximize_horizontal: true
					type: CUI.Input
					name: "system_object_id"
					form:
						label: $$("admin.event.manager.form.system_object_id")
					placeholder: $$("admin.event.manager.form.system_object_id.placeholder")
			,
				if ez5.version("6.7")
					type: ez5.ObjecttypeSelector
					form: label: $$("admin.event.manager.form.objecttype")
					name: "objecttype"
					show_name: true
					store_value: "fullname"
			,
				new CUI.Options
					maximize_horizontal: true
					placeholder: ""
					form:
						label: $$("admin.event.manager.form.user_types")
					name: "userTypes"
					options: userTypesOptions
			,
				new UserGroupSelector
					maximize_horizontal: true
					form: label: $$("admin.event.manager.form.id_user")
					name: "user"
					aclWhoManager: new AclWhoManager(who_filter: ["user"])
			,
				userGroupSelector = new UserGroupSelector
					maximize_horizontal: true
					form: label: $$("admin.event.manager.form.id_group")
					aclWhoManager: new AclWhoManager
						store_full_data: true
						filter_groups: ["system"]
						who_filter: ["group"]
					name: "group"
					data: groupData
					onDataChanged: =>
						group = groupData.group
						if not group
							return

						group = Group.newInstance(group: group)
						idGroup = group.getId()
						if CUI.util.idxInArray(null, groupOptions, ((v) -> v.value == idGroup)) == -1
							groupOptions.push(value: idGroup, text: group.getDisplayText())

						userGroupSelector.setValue()

						CUI.util.pushOntoArray(idGroup, @__filter_data.groups)
						@__filter_form.getFieldsByName("groups")[0].show(true).reload()
			,
				new CUI.Options
					maximize_horizontal: true
					placeholder: ""
					name: "groups"
					options: groupOptions
					hidden: true
			]
		.start()

		for userType in ez5.session.getDefaults().server.events?.searchable_event_user_types or []
			userTypesOptions.push(text: userType)
		userTypesOptionsField = @__filter_form.getFieldsByName("userTypes")[0]
		userTypesOptionsField.reload()
		@__empty_filter_data = CUI.util.copyObject(@__filter_data, true)
		sp = new CUI.SimplePane
			content: @__filter_form
			footer_right: =>
				buttons = [
					ui: "admin-event-manager-form-reset-button"
					loca_key: "admin.event.manager.form.button.reset"
					onClick: =>
						@__resetFilters()
						return
				,
					ui: "admin-event-manager-form-refresh-button"
					loca_key: "admin.event.manager.form.button.go"
					primary: true
					onClick: =>
						@__refresh()
						return
				]
				buttons

		# # @centerLayout.replace(super(), "left")
		return sp
		tpc = super()
		vl = new CUI.VerticalLayout
			maximize: true
			top:
				content: sp
			center:
				content: tpc
		vl

	__resetFilters: ->
		@__filter_data = CUI.util.copyObject(@__empty_filter_data, true)
		@data = null
		@load()
		return

	__refresh: ->
		wb = new CUI.WaitBlock(element: @getTree()).show()
		@loadTree(true)
		.always =>
			wb.destroy()

	# set right icon on all sort buttons
	__updateSortButtons: ->
		for sc in @__sortColumns
			if @__sort?.startsWith(sc.key)
				sort_sc = sc
			else
				sc.button.replace(new CUI.Icon(), "right") # empty placeholder icon

		if not sort_sc
			return

		if @__sort.match(/DESC$/)
			sort_sc.button.replace(new CUI.Icon(icon: "south"), "right")
		else
			sort_sc.button.replace(new CUI.Icon(icon: "north"), "right")
		@

	getHeaderColumn: (key) ->

		# Not sorting columns
		if key in ["type"]
			return new CUI.ListViewHeaderColumn(element: new LocaLabel(
				loca_key: "admin.event.manager.column."+key
				right: true
			))

		sb = new LocaButton
			loca_key: "admin.event.manager.column."+key
			appearance: "link"
			right: true
			onClick: =>
				if not @__sort or not @__sort.startsWith(key)
					@__sort = key+".DESC"
				else if @__sort.endsWith("DESC")
					@__sort = key+".ASC"
				else
					@__sort = "_id.DESC"
				@__updateSortButtons()
				@loadTree(true)

		if key == "type_key"
			key = "type"

		@__sortColumns.push
			button: sb
			key: key

		new CUI.ListViewHeaderColumn(element: sb)

	@locaType: (k) ->
		ez5.loca.str_default('admin.event.manager.event.'+k.toLowerCase(), k)

	getSelectFilter: (key, keys) ->
		opts = []

		options = keys.map((key) ->
			value: key
			text: AdminEventManager.locaType(key)
		)

		options.sort((optionOne, optionTwo) ->
			optionOne.text.localeCompare(optionTwo.text)
		)

		for option in options
			opts.push
				ui: "admin-event-manager-" + option.value + "-select-option"
				value: option.value
				content: new CUI.VerticalList
					content: [
						new CUI.Label(class: "ez5-event-option-label", text: option.text)
						new CUI.Label(text: option.value, appearance: "secondary", size: "mini")
					]


		type: CUI.FormPopover
		maximize_horizontal: true
		ui: "admin-event-manager-type-form-select"
		popover:
			class: "ez5-admin-event-manager-select"
			use_element_width_as_min_width: true
			pointer: null
			placements: ["se"]
			pane:
				padded: true
			placement: "se"

		form:
			label: $$("admin.event.manager.form.#{key}")

		button:
			class: "ez5-admin-event-manager-select cui-button--has-caret"
			text: $$("admin.event.manager.select.button")
			icon_right: "fa-angle-down"

		renderDisplayContent: =>
			if not @__filter_data[key]?.length
				return

			div = CUI.dom.element("DIV", class: "ez5-tag-holder")
			for eventType in @__filter_data[key]
				div.appendChild(new CUI.Label(class: "ez5-tag", text: AdminEventManager.locaType(eventType) + " [#{eventType}]").DOM)
			return div

		fields: [
			type: CUI.Options
			horizontal: false
			name: key
			options: opts
		]

	getMissingName: ->
		$$("admin.event.manager.missing_name")

	getLeftTitle: ->
		$$("admin.event.manager.search.title")

	getCenterTitle: ->
		$$("admin.event.manager.pane_title")

	getMenuTitle: ->
		new CUI.Label(text: $$("admin.event.manager.title"))

	getListFields: ->
		# empty

	getTitle: ->
		new CUI.Label(text: $$("admin.event.manager.pane_title"))

	loadList: (filter, opts={}) ->

		if opts.searchOpts
			searchOpts = opts.searchOpts
		else
			searchOpts = {}

		searchOpts = @__getSearchOpts(searchOpts)

		# console.debug "AdminEventManager.loadList", dump(@__filter_data), "SORT:", @__sort, filter, dump(searchOpts)

		dfr = new CUI.Deferred()

		searchCallback = (data) =>
			if CUI.util.isEmpty(data)
				dfr.reject()
				return
			# console.debug "events received", data
			data.offset ?= 0

			events = []
			for event in data.objects
				events.push new AdminEventManagerEvent
					event: event
					manager: @

			if events.length > 0
				@__csvButton.enable()
				@__deleteButton?.enable()
			else
				@__csvButton.disable()
				@__deleteButton?.disable()

			_data =
				count: data.count
				offset: data.offset
				limit: data.limit
				objects: data

			@__count = data.count

			dfr.resolve(events, _data)

		ez5.splash.show("admin.event.manager.splash.loading-events")

		if @__newApi or CUI.util.isEmpty(@__filter_data.id)
			ez5.api.event(
				api: "/list"
				data: searchOpts
			).done(searchCallback).fail(dfr.reject)
		else
			ez5.api.event(api: "/#{@__filter_data.id}").done((data) =>
				data =
					objects: [data]
					count: 1
					limit: 1
				searchCallback(data)
				@__csvButton.disable()
				@__deleteButton?.disable()
			).fail( =>
				# If the search by id fails, searchCallback is invoked with an empty array of events.
				data =
					objects: []
					count: 0
					limit: 1
				searchCallback(data)
			)

		dfr.always =>
			ez5.splash.hide()

		dfr.promise()

	__getSearchOpts: (searchOpts = {}) ->
		if not searchOpts.limit
			searchOpts.limit = 100

		if not searchOpts.offset
			searchOpts.offset = 0

		if @__newApi and not CUI.util.isEmpty(@__filter_data.id)
			searchOpts.id = @__filter_data.id

		if @__filter_data.type?.length > 0
			searchOpts.type = @__filter_data.type.join(",")

		if not CUI.util.isEmpty(@__filter_data.from)
			searchOpts.date_from = @__filter_data.from

		if not CUI.util.isEmpty(@__filter_data.to)
			searchOpts.date_to = @__filter_data.to

		if not CUI.util.isEmpty(@__filter_data.user)
			searchOpts.user_id = @__filter_data.user.user._id

		if not CUI.util.isEmpty(@__filter_data.groups)
			searchOpts.group_id = @__filter_data.groups.join(",")

		if not CUI.util.isEmpty(@__filter_data.userTypes)
			searchOpts.user_type = @__filter_data.userTypes.join(",")

		if ez5.version("6.7") and not CUI.util.isEmpty(@__filter_data.objecttype)
			searchOpts.objecttype = @__filter_data.objecttype

		if ez5.version("6.7") and not CUI.util.isEmpty(@__filter_data.system_object_id)
			searchOpts.system_object_id = @__filter_data.system_object_id

		if @__sort
			if @__sort.indexOf("_id.") == -1
				searchOpts.sort = @__sort+",_id.DESC"
			else
				searchOpts.sort = @__sort

		# console.debug "admin events", @__sort, searchOpts
		searchOpts

	loadTree: (refresh = false) ->
		if refresh
			return super()
		else
			@setStatus()
			return CUI.resolvedPromise()

	__getCSVDownloadURL: (_searchOpts) ->
		searchOpts = @__getSearchOpts()
		delete(searchOpts.limit)
		CUI.util.mergeMap(searchOpts, _searchOpts)

		searchOpts.format = "csv"
		searchOpts.csv_use_bom = "true"
		searchOpts.csv_explode_array_concat = ","
		searchOpts.csv_delimiter = ";"
		searchOpts.offset = 0

		url = ez5.api.getFullBase()+"/event/list?"+ez5.tokenParam+"="+ez5.session.token+"&"
		url += CUI.encodeUrlData(searchOpts)
		url

	__deleteAll: ->
		deleteApiInvocation = null
		deletedCount = 0

		alert = new CUI.Alert(
			text: $$("admin.event.manager.remove_loading.text")
			text_icon: "spinner"
			button_text_ok: $$("admin.event.manager.remove_loading.abort-button")
		)

		alert.open().done(=>
			deleteApiInvocation?.abort()
			alert.destroy()
			showSuccess(deletedCount)
		)

		showSuccess = (count) =>
			CUI.alert(text: $$("admin.event.manager.remove_from_server_confirm.done", count: count)).done( =>
				@__refresh()
			)

		doDelete = =>
			if @__newApi
				api = "/list"
			else
				api = ""

			deleteApiInvocation = ez5.api.event(
				type: "DELETE"
				api: api
				data: @__getDeleteUrlData()
			).done((data) =>
				if ez5.version("6")
					# On fylr we remove all the events with only one call.
					alert?.destroy()
					showSuccess( @__count)
					return

				if deletedCount >= @__count
					alert?.destroy()
					showSuccess(deletedCount)
				else
					deletedCount += data.count
					doDelete()
			).fail( =>
				alert?.destroy()
				showSuccess(deletedCount)
			)
		doDelete()


	__getDeleteUrlData: ->
		searchOpts = @__getSearchOpts()
		delete(searchOpts.limit)
		searchOpts.offset = 0
		searchOpts

	deleteRows: (rows) ->
		ids = []
		for row in rows
			ids.push(row.getID())
		@__deleteMultipleById(ids)

	load: ->
		if ez5.session.hasSystemRight("root", "system.api.event.get", "system.api.event[get]")
			loadPromise = super()
			gridBody = @__getTreeGridBody()
			if gridBody
				new CUI.Lasso
					element: gridBody
					create: (ev) =>
						ev.hasModifierKey(true)
					dragstart: =>
						selectedRows = @tree.getSelectedRows()
						if selectedRows.length > 0
							for selectedRow in selectedRows
								selectedRow?.deselect(null)
					selected: (ev, info) =>
						for element in info.elements
							listViewRow = CUI.dom.data(element, "listViewRow")
							listViewRow?.select(ev)
						return true
			return loadPromise
		else
			label = new CUI.Label
				text: $$("admin.event.manager.no_rights")
				multiline: true
				centered: true
				appearance: 'secondary'
			ez5.rootLayout.replace(label, "center")
		return CUI.resolvedPromise()

	unload: ->
		gridBody = @__getTreeGridBody()
		if gridBody
			CUI.Lasso.destroy(gridBody)
		super()

	__getTreeGridBody: ->
		grid = @tree?.getBottom()
		if not grid
			return
		return CUI.dom.children(grid)[0]

	__deleteMultipleById: (ids = []) ->

		if @__newApi
			call =
				type: "DELETE"
				api: "/list"
				data:
					id: ids.join(",")
		else
			call =
				type: "DELETE"
				json_data: ids

		if ids.length > 1
			titleKey = $$("admin.event.manager.remove_from_server_confirm.title")
			textKey = $$("admin.event.manager.remove_from_server_confirm.text", count: ids.length)
			alertKey = $$("admin.event.manager.remove_from_server_confirm.done", count: ids.length)
		else
			titleKey = $$("admin.event.manager.remove_from_server_confirm.title.one")
			textKey = $$("admin.event.manager.remove_from_server_confirm.text.one")
			alertKey = $$("admin.event.manager.remove_from_server_confirm.done.one")

		CUI.confirm
			title: titleKey
			text: textKey
		.done =>
			ez5.api.event(call).done =>
				CUI.alert(text: alertKey).done =>
					if not @__newApi and @__filter_data.id
						@__searchIdInput.setValue("")
						@__enableDisableFormFields(@__filter_data, @__searchIdInput)
					@__refresh()

	__enableDisableFormFields: (data, input) ->
		if @__newApi
			return
		for field in input.getForm().getFields()
			if field != input
				if data.id then field.disable() else field.enable()

	getPlusButton: ->


	getExtraButton: ->
		[
			if ez5.session.hasSystemRight("root", "system.api.event.delete", "system.api.event[delete]")
				@__deleteButton = new LocaButton
					ui: "admin.event.manager-remove-from-server-button"
					loca_key: "admin.event.manager.remove_from_server"
					disabled: true
					onClick: =>
						CUI.confirm
							title: $$("admin.event.manager.remove_from_server_confirm.title")
							button_text_ok: $$("admin.event.manager.remove_from_server_confirm.ok")
							button_text_cancel: $$("admin.event.manager.remove_from_server_confirm.cancel")
							text: $$("admin.event.manager.remove_from_server_confirm.text", count: @__count)
						.done =>
							@__deleteAll()
		,
			@__csvButton = new LocaButtonHref
				loca_key: "admin.event.manager.csv_download"
				disabled: true
				onClick: =>
					data =
						csv_max_length: true
						csv_explode_info: true
					form = new CUI.Form
						data: data
						fields: [
							type: CUI.Checkbox
							name: "csv_explode_info"
							form: label: $$("admin.event.manager.csv-download.form.explode_info.label")
						,
							type: CUI.Checkbox
							name: "csv_max_length"
							form: label: $$("admin.event.manager.csv-download.form.max_length.label")
						]
					form.start()

					modal = new CUI.Modal
						cancel: true
						onCancel: =>
							modal.destroy()
							return
						pane:
							padded: true
							header_left: new CUI.Label
								text: $$("admin.event.manager.csv-download.modal.header.label")
							content: form
							footer_right: new LocaButton
								loca_key: "admin.event.manager.csv-download.modal.button"
								onClick: =>
									opts =
										csv_max_length: if data.csv_max_length then 100 else 0
										csv_explode: if data.csv_explode_info then "true" else undefined
									downloadUrl = @__getCSVDownloadURL(opts)
									ez5.download_file(href: downloadUrl)
									modal.doCancel()
					modal.show()
		]

ez5.session_ready ->
	ez5.rootMenu.registerApp(AdminEventManager)
